"""
Tests for background module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)

from unittest import TestCase
import numpy as np
from numpy.testing import assert_equal, assert_almost_equal
from uncertainties import unumpy as unp
from islatu import background


class TestBackground(TestCase):
    """
    Unit tests for background module
    """

    def test_bivariate_normal(self):
        """
        A simple test for the generation of a bivariate normal distribution.
        """
        x_1 = np.linspace(-1, 1, 100)
        x_2 = np.linspace(-1, 1, 100)
        input_x = np.array([x_1, x_2])
        abscissa = np.array(np.mgrid[0 : len(input_x[0]) : 1, 0 : len(input_x[1]) : 1])

        output = background.bivariate_normal(abscissa, 0, 0, 1, 1, 1, 10)
        assert_equal(10000, output.size)
        output = output.reshape((100, 100))
        max_inten = np.unravel_index(np.argmax(output, axis=None), output.shape)
        assert_equal([0, 0], max_inten)

    def test_fit_gaussian_2d(self):
        """
        Test the fitting of a 2-d gaussian function.
        """
        x_1 = np.linspace(-1, 1, 10)
        x_2 = np.linspace(-1, 1, 10)
        input_x = np.array([x_1, x_2])
        abscissa = np.array(np.mgrid[0 : len(input_x[0]) : 1, 0 : len(input_x[1]) : 1])
        to_fit = background.bivariate_normal(abscissa, 0, 0, 1, 1, 15, 10)
        to_fit = to_fit.reshape((10, 10))
        to_fit_e = to_fit * 0.1

        result = background.fit_gaussian_2d(to_fit, to_fit_e)
        assert_almost_equal(unp.nominal_values(result[0]), [0, 0, 1, 1, 15, 10])
        assert_equal(4, result[1])



"""
Tests for refl_data module
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)

from os import path
import islatu
from unittest import TestCase
import numpy as np
from scipy.interpolate import splrep
from numpy.testing import assert_equal, assert_almost_equal
from uncertainties import unumpy as unp
from islatu import refl_data, io, background, cropping


class TestReflData(TestCase):
    """
    Unit tests for refl_data module
    """

    def test_init_profile(self):
        """
        Test init with correct file paths
        """
        file_name1 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        file_name2 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        files = [file_name1, file_name2]
        r = refl_data.Profile(files, io.i07_dat_parser)
        assert_equal(len(r.scans), 2)

    def test_crop_bkg_profile(self):
        """
        Test crop and bkg sub profile
        """
        file_name1 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        file_name2 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        files = [file_name1, file_name2]
        r = refl_data.Profile(files, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        a2 = np.zeros((3))
        np.any(np.not_equal(unp.nominal_values(r.scans[0].R), a2))
        np.any(np.not_equal(unp.nominal_values(r.scans[1].R), a2))
        np.any(np.not_equal(unp.std_devs(r.scans[0].R), a2))
        np.any(np.not_equal(unp.std_devs(r.scans[1].R), a2))
        np.any(np.not_equal(r.scans[0].n_pixels, a2))
        np.any(np.not_equal(r.scans[1].n_pixels, a2))

    def test_footprint_profile(self):
        """
        Test footprint corrections profile
        """
        file_name1 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        file_name2 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        files = [file_name1, file_name2]
        r = refl_data.Profile(files, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        r_store0 = r.scans[0].R
        r_store1 = r.scans[1].R
        r.footprint_correction(100e-6, 100e-3)
        np.any(
            np.not_equal(unp.nominal_values(r.scans[0].R), unp.nominal_values(r_store0))
        )
        np.any(
            np.not_equal(unp.nominal_values(r.scans[1].R), unp.nominal_values(r_store1))
        )

    def test_transmission_profile(self):
        """
        Test transmission correction profile
        """
        file_name1 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        file_name2 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        files = [file_name1, file_name2]
        r = refl_data.Profile(files, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        r_store0 = r.scans[0].R
        r_store1 = r.scans[1].R
        r.transmission_normalisation()
        np.any(
            np.not_equal(unp.nominal_values(r.scans[0].R), unp.nominal_values(r_store0))
        )
        np.any(
            np.not_equal(unp.nominal_values(r.scans[1].R), unp.nominal_values(r_store1))
        )

    def test_qdcd_normalisation_profile(self):
        """
        Test qdcd normalisation profile
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/qdcd_norm.dat"
        )
        normalisation_metadata, normalisation_data = io.i07_dat_parser(file_name)
        itp = splrep(normalisation_data["qdcd_"], normalisation_data["adc2"])
        file_name1 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        file_name2 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        files = [file_name1, file_name2]
        r = refl_data.Profile(files, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        r_store0 = r.scans[0].R
        r_store1 = r.scans[1].R
        r.qdcd_normalisation(itp)
        np.any(
            np.not_equal(unp.nominal_values(r.scans[0].R), unp.nominal_values(r_store0))
        )
        np.any(
            np.not_equal(unp.nominal_values(r.scans[1].R), unp.nominal_values(r_store1))
        )

    def test_resolution_function_profile(self):
        """
        Test resolution_function from pixel profile
        """
        file_name1 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        file_name2 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        files = [file_name1, file_name2]
        r = refl_data.Profile(files, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        q_store0 = unp.nominal_values(r.scans[0].q)
        q_store1 = unp.nominal_values(r.scans[1].q)
        r.resolution_function(1)
        assert_equal(unp.nominal_values(r.scans[0].q), q_store0)
        np.any(np.not_equal(unp.std_devs(r.scans[0].q), np.zeros((3))))
        assert_equal(unp.nominal_values(r.scans[1].q), q_store1)
        np.any(np.not_equal(unp.std_devs(r.scans[1].q), np.zeros((3))))

    def test_concatentate_profile(self):
        """
        Test profile concatentation
        """
        file_name1 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        file_name2 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        files = [file_name1, file_name2]
        r = refl_data.Profile(files, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        r.concatenate()
        assert_equal(r.q.size, 6)
        assert_equal(r.R.size, 6)
        assert_equal(r.dq.size, 6)
        assert_equal(r.dR.size, 6)

    def test_normalise_ter_profile(self):
        """
        Test profile ter normalisation
        """
        file_name1 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        file_name2 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        files = [file_name1, file_name2]
        r = refl_data.Profile(files, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        r.concatenate()
        store = r.R
        dstore = r.dR
        r.normalise_ter()
        assert_equal(r.q.size, 6)
        assert_equal(r.R.size, 6)
        np.any(np.not_equal(r.R, store))
        np.any(np.not_equal(r.dR, dstore))

    def test_rebin_profile(self):
        """
        Test profile rebin
        """
        file_name1 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        file_name2 = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        files = [file_name1, file_name2]
        r = refl_data.Profile(files, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        r.concatenate()
        r.rebin(number_of_q_vectors=3)
        assert_equal(r.q.size, 2)
        assert_equal(r.R.size, 2)

    def test_init_a(self):
        """
        Test init with correct file path
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser)
        assert_equal(len(r.data["file"]), 3)
        assert_equal(
            r.data["file"][0], "islatu/tests/test_files/location/pilatus1/tiff_a.tif"
        )
        assert_equal(
            r.data["file"][1], "islatu/tests/test_files/location/pilatus1/tiff_b.tif"
        )
        assert_equal(
            r.data["file"][2], "islatu/tests/test_files/location/pilatus1/tiff_c.tif"
        )

    def test_init_b(self):
        """
        Test init with -2: file path correct
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_b.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser)
        assert_equal(len(r.data["file"]), 3)
        assert_equal(
            r.data["file"][0],
            path.join(
                path.dirname(islatu.__file__), "tests/test_files/pilatus1/tiff_a.tif"
            ),
        )
        assert_equal(
            r.data["file"][1],
            path.join(
                path.dirname(islatu.__file__), "tests/test_files/pilatus1/tiff_b.tif"
            ),
        )
        assert_equal(
            r.data["file"][2],
            path.join(
                path.dirname(islatu.__file__), "tests/test_files/pilatus1/tiff_c.tif"
            ),
        )

    def test_init_c(self):
        """
        Test init with -1 file path correct
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_c.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser)
        assert_equal(len(r.data["file"]), 3)
        assert_equal(
            r.data["file"][0],
            path.join(path.dirname(islatu.__file__), "tests/test_files/tiff_a.tif"),
        )
        assert_equal(
            r.data["file"][1],
            path.join(path.dirname(islatu.__file__), "tests/test_files/tiff_b.tif"),
        )
        assert_equal(
            r.data["file"][2],
            path.join(path.dirname(islatu.__file__), "tests/test_files/tiff_c.tif"),
        )

    def test_init_e(self):
        """
        Test init with correct file path but q_axis as none
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser, q_axis_name=None)
        assert_equal(len(r.data["file"]), 3)
        assert_almost_equal(unp.nominal_values(r.q), r.data["qdcd"], decimal=5)

    def test_str(self):
        """
        Test string output
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser, q_axis_name=None)
        assert_equal(
            r.__str__(),
            "The file: {} contains 3 images from q = 0.0100 to q = 0.0110.".format(
                file_name
            ),
        )

    def test_repr(self):
        """
        Test string output
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser, q_axis_name=None)
        assert_equal(
            r.__repr__(),
            "The file: {} contains 3 images from q = 0.0100 to q = 0.0110.".format(
                file_name
            ),
        )

    def test_init_d(self):
        """
        Test init with wrong file path
        """
        with self.assertRaises(FileNotFoundError):
            file_name = path.join(
                path.dirname(islatu.__file__), "tests/test_files/test_d.dat"
            )
            r = refl_data.Scan(file_name, io.i07_dat_parser)

    def test_crop_bkg(self):
        """
        Test crop and bkg sub
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        a2 = np.zeros((3))
        np.any(np.not_equal(unp.nominal_values(r.R), a2))
        np.any(np.not_equal(unp.std_devs(r.R), a2))
        np.any(np.not_equal(r.n_pixels, a2))

    def test_crop_bkg_kwargs(self):
        """
        Test crop and bkg sub with kwargs
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser)
        p0 = [
            5,
            5,
            1,
            1,
            1,
            1000,
        ]
        r.crop(
            cropping.crop_around_peak_2d,
            kwargs={"x_size": 20},
        )
        r.bkg_sub(
            background.fit_gaussian_2d,
            kwargs={"p0": p0},
        )
        a2 = np.zeros((3))
        np.any(np.not_equal(unp.nominal_values(r.R), a2))
        np.any(np.not_equal(unp.std_devs(r.R), a2))
        np.any(np.not_equal(r.n_pixels, a2))

    def test_footprint(self):
        """
        Test footprint corrections
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        r_store = r.R
        r.footprint_correction(100e-6, 100e-3)
        np.any(np.not_equal(unp.nominal_values(r.R), unp.nominal_values(r_store)))

    def test_transmission(self):
        """
        Test transmission correction
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        r_store = r.R
        r.transmission_normalisation()
        np.any(np.not_equal(unp.nominal_values(r.R), unp.nominal_values(r_store)))

    def test_qdcd_normalisation(self):
        """
        Test qdcd normalisation
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/qdcd_norm.dat"
        )
        normalisation_metadata, normalisation_data = io.i07_dat_parser(file_name)
        itp = splrep(normalisation_data["qdcd_"], normalisation_data["adc2"])
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        r_store = r.R
        r.qdcd_normalisation(itp)
        np.any(np.not_equal(unp.nominal_values(r.R), unp.nominal_values(r_store)))

    def test_resolution_function(self):
        """
        Test resolution_function from pixel
        """
        file_name = path.join(
            path.dirname(islatu.__file__), "tests/test_files/test_a.dat"
        )
        r = refl_data.Scan(file_name, io.i07_dat_parser)
        r.crop(cropping.crop_around_peak_2d)
        r.bkg_sub(background.fit_gaussian_2d)
        q_store = unp.nominal_values(r.q)
        r.resolution_function(1)
        assert_equal(unp.nominal_values(r.q), q_store)
        np.any(np.not_equal(unp.std_devs(r.q), np.zeros((3))))



"""
The two-dimension detector generates images of the reflected intensity.
The purpose of this class is the investigation and manipulation of these
images.
"""

# Copyright (c) Andrew R. McCluskey
# Distributed under the terms of the MIT License
# author: Andrew R. McCluskey (andrew.mccluskey@diamond.ac.uk)

import numpy as np
from PIL import Image as PILIm
from uncertainties import unumpy as unp

from islatu.background import fit_gaussian_1d


class Image:
    """
    This class stores information about the detector images.

    Attributes:
        file_path (:py:attr:`str`): File path for the image.
        data (:py:class:`pandas.DataFrame`): Experimental data about the
            measurement.
        metadata (:py:attr:`dict`): Metadata regarding the measurement.
        array (:py:attr:`array_like`): The image described as an array.
        bkg (:py:class:`uncertainties.cores.Variable`): The background that
            was subtracted from the image.
        n_pixels (:py:attr:`float`): The width of the peak in number of
            pixels, used to calculate an uncertainty in q on the detector.

    Args:
        file_path (:py:attr:`str`): The file path for the image.
        data (:py:class:`pandas.DataFrame`, optional): Experimental data about
            the measurement. Defaults to :py:attr:`None`.
        metadata (:py:attr:`dict`, optional): Metadata regarding the
            measurement. Defaults to :py:attr:`None`.
        transpose (:py:attr:`bool`, optional): Should the data be rotated by
            90 degrees? Defaults to :py:attr:`False`.
        pixel_min (:py:attr:`int`): The minimum pixel value possible, all
            pixels found with less than this value will have this value
            assigned. Defaults to :py:attr:`0`.
        hot_pixel_max (:py:attr:`int`): The value of a hot pixel that will be
            set to the average of the surrounding pixels. Logically, this
            should be less than the :py:attr:`pixel_max` value. Defaults to
            :py:attr:`2e5`.
    """
    def __init__(self, file_path, data=None, metadata=None, transpose=False,
                 pixel_min=0, hot_pixel_max=2e5):
        """
        Initialisation of the :py:class:`islatu.image.Image` class, includes
        assigning uncertainties.
        """
        self.file_path = file_path
        self.data = data
        self.metadata = metadata
        img = PILIm.open(file_path)
        array = np.array(img)
        img.close()
        if transpose:
            array = array.T
        array = _average_out_hot(array, hot_pixel_max)
        array[np.where(array < pixel_min)] = 0
        array_error = np.sqrt(array)
        array_error[np.where(array == 0)] = 1
        self.array = unp.uarray(array, array_error)
        self.bkg = 0
        self.n_pixel = 0

    @property
    def nominal_values(self):
        """
        Get the nominal values of the image array.

        Returns:
            :py:attr:`array_like`: Nominal values of image.
        """
        return unp.nominal_values(self.array)

    @property
    def std_devs(self):
        """
        Get the standard deviation values of the image array.

        Returns:
            :py:attr:`array_like`: Standard deviation values of image.
        """
        return unp.std_devs(self.array)

    @property
    def n(self):
        """
        Get the nominal values of the image array.

        Returns:
            :py:attr:`array_like`: Nominal values of image.
        """
        return unp.nominal_values(self.array)

    @property
    def s(self):
        """
        Get the standard deviation values of the image array.

        Returns:
            :py:attr:`array_like`: Standard deviation values of image.
        """
        return unp.std_devs(self.array)

    @property
    def shape(self):
        """
        Array shape

        Returns:
            :py:attr:`tuple` of :py:attr:`int`: The shape of the image.
        """
        return unp.nominal_values(self.array).shape

    def __repr__(self):
        """
        Custom representation.

        Returns:
            :py:attr:`array_like`: Image array.
        """
        return self.array

    def __str__(self):
        """
        Custom string.

        Returns:
            :py:attr:`array_like`: Image array.
        """
        return self.array

    def crop(self, crop_function, **kwargs):
        """
        Perform an image crop based on some function.

        Args:
            crop_function (:py:attr:`callable`): The function to crop the data.
            **kwargs (:py:attr:`dict`): The crop function keyword arguments.
        """
        self.array = unp.uarray(*crop_function(self.n, self.s, **kwargs))

    def background_subtraction(self, background_subtraction_function,
                               **kwargs):
        """
        Perform a background subtraction based on some function.

        Args:
            background_subtraction_function (:py:attr:`callable`): The
                function to model the data and therefore remove the background.
            **kwargs (:py:attr:`dict`): The background substraction function
                keyword arguments.
        """
        bkg_popt, bkg_idx = background_subtraction_function(
            self.n, self.s, **kwargs
        )
        self.bkg = bkg_popt[bkg_idx]
        self.array -= self.bkg

    def sum(self, axis=None):
        """
        Perform a summation on the image

        Args:
            axis (:py:attr:`int`, optional): The axis of the array to perform
                the summation over. Defaults to :py:attr:`None`.
        """
        return self.array.sum(axis)

    def q_resolution(self, qz_dimension=1):
        """
        Estimate the q-resolution function based on the reflected intensity
        on the detector.

        Args:
            qz_dimension (:py:attr:`int`, optional): The dimension of q_z in
                the detector image (this should be the opposite index to that
                the summation is performed if the
                :py:func:`islatu.background.fit_gaussian_1d` background
                subtraction has been performed). Defaults to :py:attr:`1`.
        """
        bkg_popt, bkg_idx = fit_gaussian_1d(
            self.n, self.s, axis=qz_dimension
        )
        self.n_pixel = bkg_popt[1]


def _average_out_hot(array, hot_pixel_max=2e5):
    """
    Make hot pixels equal to the local average.

    Args:
        array (:py:attr:`array_like`): The array to have hot pixels removed.
        hot_pixel_max (:py:attr:`float`, optional): The definition of a hot
            pixel.

    Returns:
       :py:attr:`array_like`: Image array with hot pixels removed.
    """
    x, y = np.where(array > hot_pixel_max)
    for i in range(x.size):
        a = x[i]
        b = y[i]
        j = 1
        k = 1
        if a == 0:
            a = 1
            j = 0
        elif a == array.shape[0] - 1:
            a = array.shape[0] - 1
            j = 1
        if b == 0:
            b = 1
            k = 0
        elif b == array.shape[1] - 1:
            b = array.shape[1] - 1
            k = 1
        local = np.copy(array[a-1:a+2, b-1:b+2])
        local[j, k] = 0
        if local.mean() < (array[x[i], y[i]]) / 100:
            array[x[i], y[i]] = local.sum() / (local.size - 1)
    return array
